/*
 * Copyright (C) 2005 by egnite Software GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*!
 * \file parport.c
 * \brief Hardware dependant part of the physical JTAG layer.
 *
 * This file is used for the Windows PC parallel port and requires the
 * installation of the giveio.sys driver.
 *
 * \verbatim
 *
 * $Log: parport.c,v $
 * Revision 1.1.1.1  2005/09/14 09:01:09  haraldkipp
 * Initial import.
 *
 *
 * \endverbatim
 */

#include <conio.h>
#include <windows.h>

#include "tap.h"

#define PORTDRIVER  "\\\\.\\giveio"

#define PORT_RST_MASK   0x01    /* Pin 2 */
#define PORT_TMS_MASK   0x02    /* Pin 3 */
#define PORT_TCK_MASK   0x04    /* Pin 4 */
#define PORT_TDI_MASK   0x08    /* Pin 5 */
#define PORT_TDO_MASK   0x80    /* Pin 11 */

static int port_shadow = PORT_RST_MASK;
static unsigned short port_base;

/*!
 * \brief Set TDI bit.
 *
 * \param s Specifies wether TDI should be set (1) or cleared (0).
 */
void TapTdi(int s)
{
    if (s) {
        port_shadow |= PORT_TDI_MASK;
    } else {
        port_shadow &= ~PORT_TDI_MASK;
    }
}

/*!
 * \brief Get TDO bit.
 *
 * \return 1 if TDI is set, 0 otherwise.
 */
int TapTdo(void)
{
    if (_inp((unsigned short) (port_base + 1)) & PORT_TDO_MASK) {
        return 0;
    }
    return 1;
}

/*!
 * \brief Go to next TAP state.
 *
 * \param s Specifies wether TMS should be set (1) or cleared (0).
 */
void TapNext(int s)
{
    if (s) {
        port_shadow |= PORT_TMS_MASK;
    } else {
        port_shadow &= ~PORT_TMS_MASK;
    }

    /* Rising edge at TCK. */
    _outp(port_base, port_shadow);
    _outp(port_base, port_shadow | PORT_TCK_MASK);
}

/*!
 * \brief Reset TAP controller.
 *
 * \param s Specifies wether TRST should be set (1) or cleared (0).
 */
void TapReset(int s)
{
    if (s) {
        _outp(port_base, 0);
    } else {
        _outp(port_base, PORT_RST_MASK);
    }
}

/*!
 * \brief Initialize TAP controller.
 *
 * \param devname Name of the device, eg. lpt1, lpt2 or lpt3.
 *
 * \return 0 on success, -1 otherwise.
 *
 * \todo Implementing the Linux version.
 */
int TapInit(char *devname)
{
    HANDLE h;

    if (*devname == '\0' || stricmp(devname, "lpt1") == 0) {
        port_base = 0x378;
    } else if (stricmp(devname, "lpt2") == 0) {
        port_base = 0x278;
    } else if (stricmp(devname, "lpt3") == 0) {
        port_base = 0x3BC;
    } else {
        return -1;
    }

    h = CreateFile(PORTDRIVER, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (h != INVALID_HANDLE_VALUE) {
        CloseHandle(h);
        TapIdle();
        return 0;
    }
    return -1;
}

/*!
 * \brief De-initialize the TAP controller.
 */
void TapExit(void)
{
    TapReset(0);
}
