/*
 * Copyright (C) 2005 by egnite Software GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*!
 * \file sernix.c
 * \brief Serial port I/O routines for Linux.
 *
 * \verbatim
 *
 * $Log: sernix.c,v $
 * Revision 1.1  2005/11/25 13:39:39  haraldkipp
 * First check-in
 *
 *
 * \endverbatim
 */

#include <ctype.h>
#include <stdio.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/select.h>
#include <fcntl.h>

#include "serport.h"

int hdev = -1;

int SerialPortOpen(char *name)
{
    struct termios opt;

    if ((hdev = open(name, O_RDWR | O_NOCTTY)) < 0) {
        return -1;
    }

    /* Get current port options. */
    if (!isatty(hdev) || tcgetattr(hdev, &opt) < 0) {
        SerialPortClose();
        return -1;
    }

    /* Raw input, raw output, 8N1, 115200 Bd. */
    opt.c_iflag = 0;
    opt.c_lflag = 0;
    opt.c_oflag = 0;
    opt.c_cflag = CS8 | CREAD | CLOCAL;
    opt.c_cc[VMIN] = 1;
    opt.c_cc[VTIME] = 0;
    cfsetospeed(&opt, B115200);
    cfsetispeed(&opt, B115200);

    /* Set new port options. */
    if (tcsetattr(hdev, TCSANOW, &opt) < 0) {
        SerialPortClose();
        return -1;
    }
    return 0;
}

void SerialPortClose(void)
{
    if (hdev >= 0) {
        close(hdev);
        hdev = -1;
    }
}

/*!
 * \return Number of bytes received, 0 on timeout or -1 if an error occured.
 */
int SerialPortRead(unsigned char *buf, int siz)
{
    int rc;
    fd_set fds;
    struct timeval tmo;

    if (hdev < 0) {
        return -1;
    }

    FD_ZERO(&fds);
    FD_SET(hdev, &fds);

    tmo.tv_sec = 10;
    tmo.tv_usec = 0;

    if ((rc = select(hdev + 1, &fds, NULL, NULL, &tmo)) <= 0) {
        return rc;
    }
    return read(hdev, buf, siz);
}

/*!
 * \return 0 if successful or -1 in case of an error.
 */
int SerialPortWrite(const unsigned char *buf, int len)
{
    if (hdev < 0) {
        return -1;
    }
    if (write(hdev, buf, len) != len) {
        return -1;
    }
    return 0;
}

void SerialPortClear(void)
{
    int rc;
    fd_set fds;
    struct timeval tmo;
    char buf[16];

    if (hdev < 0) {
        return;
    }

    FD_ZERO(&fds);
    FD_SET(hdev, &fds);

    tmo.tv_sec = 0;
    tmo.tv_usec = 100000;

    for (;;) {
        if (select(hdev + 1, &fds, NULL, NULL, &tmo) <= 0) {
            break;
        }
        read(hdev, buf, sizeof(buf));
    }
}
