/*
 * Copyright (C) 2005 by egnite Software GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*!
 * \file turtelizer.c
 * \brief Turtelizer Programming Adapter.
 *
 * \verbatim
 *
 * $Log: turtelizer.c,v $
 * Revision 1.3  2006/01/09 09:03:39  haraldkipp
 * Programming adapter firmware update added.
 *
 * Revision 1.2  2005/11/25 13:40:35  haraldkipp
 * Can compile on Linux now.
 *
 * Revision 1.1  2005/11/25 09:06:48  haraldkipp
 * First check-in
 *
 *
 * \endverbatim
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef _MSC_VER
#else
#define stricmp strcasecmp
#endif

#include "jtagomat.h"
#include "serport.h"

#define JTAG_OPEN       'O'
#define JTAG_CLOSE      'C'
#define JTAG_RESET      'Z'
#define JTAG_ENTER      'E'
#define JTAG_EXIT       'X'
#define JTAG_ID         'I'
#define JTAG_REGREAD    'r'
#define JTAG_REGWRITE   'w'
#define JTAG_MEMREAD    'R'
#define JTAG_MEMWRITE   'W'
#define JTAG_COMREAD    'i'
#define JTAG_COMWRITE   'o'

#define FIRMWARE_START  '0'
#define FIRMWARE_ERASE  '1'
#define FIRMWARE_WRITE  '2'
#define FIRMWARE_READ   '3'
#define FIRMWARE_END    '4'

#define IERR_BAD_RESPONSE   301
#define IERR_NO_RESPONSE    302
#define IERR_RECEIVE        303
#define IERR_TRANSMIT       304

#define IERR_BAD_REGNAME    401

#define ASC_STX     0x02
#define ASC_ETX     0x03
#define ASC_ACK     0x06
#define ASC_DLE     0x10
#define ASC_NAK     0x15

#define MAX_TELE_SIZE   256

static int ice_error;

/*!
 * \brief Return index of a given register name.
 *
 * \param name Name of the register.
 *
 * \return Register index or -1 if the name is undefined.
 */
static int Arm7RegIndex(char *name)
{
    int rc;

    if (toupper(name[0]) == 'R' && isdigit(name[1])) {
        rc = atoi(name + 1);
        if (rc < 0 || rc > 15) {
            rc = -1;
        }
    } else if (stricmp(name, "SP") == 0) {
        rc = 13;
    } else if (stricmp(name, "LR") == 0) {
        rc = 14;
    } else if (stricmp(name, "PC") == 0) {
        rc = 15;
    } else if (stricmp(name, "CPSR") == 0) {
        rc = 16;
    } else {
        rc = -1;
    }
    return rc;
}


/*!
 * \return Number of bytes received, 0 on timeout or -1 if an error occured.
 */
static int IceTeleRecv(unsigned char *buf, int siz)
{
    int retry = 3;
    unsigned char ch;
    unsigned char cs = 0;
    unsigned char st = 0;
    int rc = 0;
    unsigned char *tbuf;
    int tpos;
    int tlen;

    /* Allocate receiver buffer. */
    if ((tbuf = malloc(MAX_TELE_SIZE)) == NULL) {
        return -1;
    }
    /* Reset current read position. */
    tpos = 0;
    /* Reset current content. */
    tlen = 0;

    /*
     * Loop for each character until we got a complete telegram.
     */
    for (;;) {
        /* If our read buffer is empty, try to get more from the line. */
        if (tpos == tlen) {
            if ((tlen = SerialPortRead(tbuf, MAX_TELE_SIZE)) == -1) {
                /* Fatal receiver error. */
                rc = -1;
                break;
            }
            if (tlen == 0) {
                /* Handle timeout. */
                rc = 0;
                if (st && retry) {
                    /*
                     * If the timeout occured in the middle of a telegram,
                     * we may have lost some data. Respond with NAK in the
                     * hope that the remote will repeat the transmission.
                     */
                    retry--;
                    st = 0;
                    ch = ASC_NAK;
                    if (SerialPortWrite(&ch, 1)) {
                        /* Fatal transmitter error. */
                        rc = -1;
                        break;
                    }
                    continue;
                }
                /* Nothing received or no more retries. Return with timeout. */
                break;
            }
            /*
             * If we reached this point, then there is new data
             * in our receive buffer.
             */
            tpos = 0;
        }
        ch = tbuf[tpos];
        tpos++;
        cs ^= ch;

        /*
         * State 1: Receiving data.
         */
        if (st == 1) {
            /* DLE is used to escape control codes. */
            if (ch == ASC_DLE) {
                st = 2;
                cs ^= ch;
                continue;
            }
            else if (ch == ASC_ETX) {
                st = 3;
                continue;
            }
            /* We do not expect STX without DLE in the data block. */
            else if (ch == ASC_STX) {
                st = 0;
            }
        }

        /*
         * State 0: Waiting for STX.
         */
        if (st == 0) {
            if (ch == ASC_STX) {
                cs = 0;
                rc = 0;
                st++;
            }
            continue;
        }

        /*
         * State 3: Verify checksum.
         */
        if (st == 3) {
            if (cs) {
                ch = ASC_NAK;
                if (retry--) {
                    st = 0;
                }
                else {
                    rc = -1;
                }
            }
            else {
                ch = ASC_ACK;
            }
            if (SerialPortWrite(&ch, 1)) {
                rc = -1;
            }
            if (ch == ASC_ACK || rc == -1) {
                break;
            }
            continue;
        }

        /*
         * State 1/2: Receiving data.
         * State 2 indicates, that the last character was DLE.
         */
        if (rc < siz) {
            buf[rc] = ch;
        }
        rc++;
        st = 1;
    }

    /* Clean up. */
    free(tbuf);

    /* Check for overflow. */
    if (rc > siz) {
        rc = -1;
    }
    return rc;
}

/*!
 * \return 0 if successful or -1 in case of an error.
 */
static int IceTeleSend(const char *buf, int len)
{
    unsigned char ch;
    unsigned char cs = 0;
    int rc = 0;
    const char *bp;
    char *tbuf;
    char *tp;
    int tlen;
    int i;

    tlen = len + 3;
    for (i = 0, bp = buf; i < len; i++, bp++) {
        if (*bp == ASC_STX || *bp == ASC_ETX || *bp == ASC_DLE) {
            tlen++;
        }
    }

    /*
     * Allocate a buffer and fill it with a complete telegram.
     */
    tbuf = malloc(tlen);
    bp = buf;
    tp = tbuf;
    *tp++ = ASC_STX;
    for (i = 0; i < len; i++) {
        cs ^= *bp;
        if (*bp == ASC_STX || *bp == ASC_ETX || *bp == ASC_DLE) {
            *tp++ = ASC_DLE;
        }
        *tp++ = *bp++;
    }
    *tp++ = ASC_ETX;
    *tp++ = cs ^ ASC_ETX;

    for (i = 0; i < 3; i++) {
        SerialPortClear();
        /* Send the telegram. */
        if ((rc = SerialPortWrite(tbuf, tlen)) != 0) {
            break;
        }

        /* Receive acknowledge. */
        if ((rc = SerialPortRead(&ch, 1)) < 0) {
            break;
        }
        if (rc = 1 && ch == ASC_ACK) {
            rc = 0;
            break;
        }
        rc = -1;
    }
    free(tbuf);

    return rc;
}

/*!
 * \brief Return the last error code.
 *
 * This function will clear the error.
 *
 * \return Error code.
 */
int IceError(void)
{
    int rc = ice_error;

    ice_error = 0;

    return rc;
}

/*!
 * \brief Send adapter command and receive response.
 *
 * \param tbuf Pointer to a buffer containing the command to transmit.
 * \param tlen Number of bytes to transmit.
 * \param rbuf Pointer to the buffer that receives the response.
 * \param tlen Number of bytes to receive.
 *
 * \return 0 on success. -1 indicates an error. Call IceError()
 *         to retrieve the specific error code.
 */
int IceTransact(char *tbuf, int tlen, char *rbuf, int rsiz)
{
    int rc;

    if ((rc = IceTeleSend(tbuf, tlen)) == 0) {
        if ((rc = IceTeleRecv(rbuf, rsiz)) > 0) {
            if (rbuf[0] == 0) {
                if (rc == rsiz) {
                    rc = 0;
                }
                else {
                    ice_error = IERR_BAD_RESPONSE;
                    rc = -1;
                }
            }
            else {
                ice_error = rbuf[0];
                rc = -1;
            }
        }
        else if (rc == 0) {
            ice_error = IERR_NO_RESPONSE;
            rc = -1;
        }
        else {
            ice_error = IERR_RECEIVE;
        }
    }
    else {
        ice_error = IERR_TRANSMIT;
    }
    return rc;
}

/*!
 * \brief Initialze the ICE routines.
 *
 * If in debug state, this routine will save the CPU context.
 */
void IceInit(void)
{
    unsigned char cmd[1];
    unsigned char rsp[5];

    cmd[0] = JTAG_OPEN;
    IceTransact(cmd, 1, rsp, 5);
    if (verbose) {
        fprintf(stderr, "Turtelizer %u.%u.%u\n", rsp[4], rsp[3], rsp[2]);
    }
}

int TapInit(char *devname)
{
    if (devname == NULL || *devname == '\0') {
        devname = "COM1";
    }
    return SerialPortOpen(devname);
}

/*!
 * \brief De-initialze the ICE routines.
 *
 * If in debug state, this routine will re-store the CPU context.
 */
void IceExit(void)
{
    unsigned char cmd[1];
    unsigned char rsp[1];

    cmd[0] = JTAG_CLOSE;
    IceTransact(cmd, 1, rsp, 1);
}

void TapExit(void)
{
    SerialPortClose();
}

/*!
 *
 */
void IceReset(void)
{
    unsigned char cmd[1];
    unsigned char rsp[1];

    cmd[0] = JTAG_RESET;
    IceTransact(cmd, 1, rsp, 1);
}

void TapReset(int s)
{
}

void TapIdle(void)
{
}

/*!
 * - Command: JTAG_ID(1)
 * - Result: result(1) data(4)
 */
unsigned long IceIdCode(void)
{
    unsigned long id = 0;
    unsigned char cmd[1];
    unsigned char rsp[5];

    cmd[0] = JTAG_ID;
    if (IceTransact(cmd, 1, rsp, 5) == 0) {
        memcpy(&id, &rsp[1], 4);
    }
    return id;
}

/*!
 *
 */
int IceEnterDebug(void)
{
    unsigned char cmd[1];
    unsigned char rsp[1];

    cmd[0] = JTAG_ENTER;
    return IceTransact(cmd, 1, rsp, 1);
}

/*!
 *
 */
int IceExitDebug(void)
{
    unsigned char cmd[1];
    unsigned char rsp[1];

    cmd[0] = JTAG_EXIT;
    return IceTransact(cmd, 1, rsp, 1);
}

/*!
 * - Command: JTAG_REGREAD(1) index(1)
 * - Result: result(1) data(4)
 */
int IceCpuRegRead(char *name, unsigned long *data)
{
    int rc;
    unsigned char cmd[2];
    unsigned char rsp[5];

    cmd[0] = JTAG_REGREAD;
    if ((rc = Arm7RegIndex(name)) != -1) {
        cmd[1] = (unsigned char)rc;
        if ((rc = IceTransact(cmd, 2, rsp, 5)) == 0) {
            memcpy(data, &rsp[1], 4);
        }
    }
    else {
        ice_error = IERR_BAD_REGNAME;
    }
    return rc;
}

/*!
 * - Command: JTAG_REGWRITE(1) index(1) value(4)
 * - Result: result(1)
 */
int IceCpuRegWrite(char *name, unsigned long data)
{
    int rc;
    unsigned char cmd[6];
    unsigned char rsp[1];

    cmd[0] = JTAG_REGWRITE;
    if ((rc = Arm7RegIndex(name)) != -1) {
        cmd[1] = (unsigned char)rc;
        memcpy(&cmd[2], &data, 4);
        rc = IceTransact(cmd, 6, rsp, 1);
    }
    else {
        ice_error = IERR_BAD_REGNAME;
    }
    return rc;
}

/*!
 * - Command: JTAG_MEMREAD(1) addr(4)
 * - Result: result(1) data(4)
 */
int IceMemoryRead(unsigned long addr, unsigned long *data)
{
    int rc;
    unsigned char cmd[5];
    unsigned char rsp[5];

    cmd[0] = JTAG_MEMREAD;
    memcpy(&cmd[1], &addr, 4);
    if ((rc = IceTransact(cmd, 5, rsp, 5)) == 0) {
        memcpy(data, &rsp[1], 4);
        rc = 0;
    }
    else {
        rc = -1;
    }
    return rc;
}

/*!
 * - Command: JTAG_MEMWRITE(1) addr(4) data(4)
 * - Result: result(1)
 */
int IceMemoryWrite(unsigned long addr, unsigned long data)
{
    unsigned char cmd[9];
    unsigned char rsp[1];

    cmd[0] = JTAG_MEMWRITE;
    memcpy(&cmd[1], &addr, 4);
    memcpy(&cmd[5], &data, 4);
    return IceTransact(cmd, 9, rsp, 1);
}

/*!
 * - Command: JTAG_COMREAD(1)
 * - Result: result(1) data(4)
 */
int IceComRead(unsigned long *data)
{
    int rc;
    unsigned char cmd[1];
    unsigned char rsp[5];

    cmd[0] = JTAG_COMREAD;
    if ((rc = IceTransact(cmd, 1, rsp, 5)) == 0) {
        memcpy(data, &rsp[1], 4);
    }
    return rc;
}

/*!
 * - Command: JTAG_COMWRITE(1) data(4)
 * - Result: JTAG_OK(1)
 */
int IceComWrite(unsigned long data)
{
    unsigned char cmd[5];
    unsigned char rsp[1];

    cmd[0] = JTAG_COMWRITE;
    memcpy(&cmd[1], &data, 4);
    return IceTransact(cmd, 5, rsp, 1);
}

/*!
 *
 */
void IcePrintStatus(FILE * fp)
{
}

int IceUpdate(unsigned char *image, unsigned long len)
{
    unsigned char rsp[MAX_TELE_SIZE];
    unsigned char req[MAX_TELE_SIZE];
    unsigned long pgsiz;
    unsigned long addr;
    unsigned long i;
    int rc;

    /*
     * Activate the boot loader if not already running.
     */
    req[0] = FIRMWARE_START;
    if (IceTransact(req, 1, rsp, 5)) {
        return -1;
    }
    if (verbose) {
        fprintf(stderr, "TurtleBoot %u.%u.%u\n", rsp[4], rsp[3], rsp[2]);
    }

    /*
     * Read the first page. This serves two purposes. First
     * we can check that the bootloader is working. Second
     * we can find out the page size, because the bootloader
     * responds with full pages.
     */
    req[0] = FIRMWARE_READ;
    memset(&req[1], 0, 4);
    if (IceTeleSend(req, 5)) {
        return -1;
    }
    if ((rc = IceTeleRecv(rsp, sizeof(rsp))) <= 0) {
        return -1;
    }
    if (rc < 5 || rsp[0]) {
        /* Bad response. */
        return -1;
    }
    memcpy(&pgsiz, &rsp[1], 4);
    if (verbose) {
        fprintf(stderr, "Page size: %lu Bytes\n", pgsiz);
    }

    /*
     * Erase all pages we are going to write.
     */
    for (addr = 0; addr < len; addr += pgsiz) {
        req[0] = FIRMWARE_ERASE;
        memcpy(&req[1], &addr, 4);
        if (IceTransact(req, 5, rsp, 1)) {
            return -1;
        }
    }

    /*
     * Write the firmware page by page.
     */
    while(addr) {
        addr -= pgsiz;
        if (verbose) {
            fprintf(stderr, "Page %08lX ", addr);
        }
        req[0] = FIRMWARE_WRITE;
        memcpy(&req[1], &addr, 4);
        memset(&req[5], 0xFF, pgsiz);
        for (i = 0; i < pgsiz; i++) {
            if (i > len) {
                req[i + 5] = 0xFF;
            }
            else {
                req[i + 5] = image[addr + i];
            }
        }
        if (IceTransact(req, pgsiz + 5, rsp, 1)) {
            return -1;
        }
        req[0] = FIRMWARE_READ;
        memcpy(&req[1], &addr, 4);
        if (IceTransact(req, 5, rsp, pgsiz + 5)) {
            return -1;
        }
        if (memcmp(&req[5], &rsp[5], pgsiz)) {
            return -1;
        }
        if (verbose) {
            fprintf(stderr, "OK\n");
        }
    }

    /*
     * Activate the new firmware. We do not expect a response.
     */
    req[0] = FIRMWARE_END;
    if (IceTeleSend(req, 1)) {
        return -1;
    }
    return 0;
}
